@@ -1,74 +1,49 @@
 /*
- * This file is part of the KFileMetaData project
- * Copyright (C) 2016  Varun Joshi <varunj.1011@gmail.com>
- * Copyright (C) 2016  Vishesh Handa <me@vhanda.in>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) version 3, or any
- * later version accepted by the membership of KDE e.V. (or its
- * successor approved by the membership of KDE e.V.), which shall
- * act as a proxy defined in Section 6 of version 3 of the license.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library.  If not, see <http://www.gnu.org/licenses/>.
- *
- */
+    This file is part of the KFileMetaData project
+    SPDX-FileCopyrightText: 2016 Varun Joshi <varunj.1011@gmail.com>
+    SPDX-FileCopyrightText: 2016 Vishesh Handa <me@vhanda.in>
 
+    SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
+*/
+
+#include "writercollection.h"
+#include "writer.h"
+#include "writer_p.h"
+#include "writerplugin.h"
+#include "externalwriter.h"
 #include "kfilemetadata_debug.h"
+#include "config-kfilemetadata.h"
 
 #include <QCoreApplication>
 #include <QPluginLoader>
 #include <QDir>
-
-#include <KLocalizedString>
-
-#include "writer.h"
-#include "writer_p.h"
-#include "writerplugin.h"
-#include "writercollection.h"
-#include "externalwriter.h"
-
-#include "config-kfilemetadata.h"
+#include <QMimeDatabase>
+#include <vector>
 
 using namespace KFileMetaData;
 
 class Q_DECL_HIDDEN WriterCollection::WriterCollectionPrivate
 {
 public:
-    QHash<QString, Writer*> m_writers;
+    QMultiHash<QString, Writer*> m_writers;
 
-    QList<Writer*> allWriters() const;
+    std::vector<Writer> m_allWriters;
+
+    void findWriters();
 };
 
 WriterCollection::WriterCollection()
-    : d_ptr(new WriterCollectionPrivate)
+    : d(new WriterCollectionPrivate)
 {
-    Q_D(WriterCollection);
-    const QList<Writer*> all = d->allWriters();
-
-    foreach (Writer* writer, all) {
-        foreach (const QString& type, writer->mimetypes()) {
-            d->m_writers.insertMulti(type, writer);
-        }
-    }
+    d->findWriters();
 }
 
 WriterCollection::~WriterCollection()
 {
-    Q_D(WriterCollection);
-    qDeleteAll(d->m_writers.begin(), d->m_writers.end());
     delete d;
 }
 
-
-QList<Writer*> WriterCollection::WriterCollectionPrivate::allWriters() const
+void WriterCollection::WriterCollectionPrivate::findWriters()
 {
     QStringList plugins;
     QStringList pluginPaths;
@@ -76,7 +51,7 @@ QList<Writer*> WriterCollection::WriterCollectionPrivate::allWriters() const
     QStringList externalPluginPaths;
 
     const QStringList paths = QCoreApplication::libraryPaths();
-    Q_FOREACH (const QString& libraryPath, paths) {
+    for (const QString& libraryPath : paths) {
         QString path(libraryPath + QStringLiteral("/kf5/kfilemetadata/writers"));
         QDir dir(path);
 
@@ -85,7 +60,7 @@ QList<Writer*> WriterCollection::WriterCollectionPrivate::allWriters() const
         }
 
         const QStringList entryList = dir.entryList(QDir::Files | QDir::NoDotAndDotDot);
-        Q_FOREACH (const QString& fileName, entryList) {
+        for (const QString& fileName : entryList) {
             // Make sure the same plugin is not loaded twice, even if it
             // installed in two different locations
             if (plugins.contains(fileName))
@@ -109,13 +84,12 @@ QList<Writer*> WriterCollection::WriterCollectionPrivate::allWriters() const
     }
     externalPlugins.clear();
 
-    QList<Writer*> writers;
-    Q_FOREACH (const QString& pluginPath, pluginPaths) {
+    for (const QString& pluginPath : qAsConst(pluginPaths)) {
         QPluginLoader loader(pluginPath);
 
         if (!loader.load()) {
-            qWarning() << "Could not create Writer: " << pluginPath;
-            qWarning() << loader.errorString();
+            qCWarning(KFILEMETADATA_LOG) << "Could not create Writer: " << pluginPath;
+            qCWarning(KFILEMETADATA_LOG) << loader.errorString();
             continue;
         }
 
@@ -123,10 +97,11 @@ QList<Writer*> WriterCollection::WriterCollectionPrivate::allWriters() const
         if (obj) {
             WriterPlugin* plugin = qobject_cast<WriterPlugin*>(obj);
             if (plugin) {
-                Writer* writer = new Writer;
-                writer->d_ptr->m_plugin = plugin;
+                Writer writer;
+                writer.d->m_plugin = plugin;
+                writer.setAutoDeletePlugin(Writer::DoNotDeletePlugin);
 
-                writers << writer;
+                m_allWriters.push_back(std::move(writer));
             } else {
                 qCDebug(KFILEMETADATA_LOG) << "Plugin could not be converted to a WriterPlugin";
                 qCDebug(KFILEMETADATA_LOG) << pluginPath;
@@ -137,26 +112,43 @@ QList<Writer*> WriterCollection::WriterCollectionPrivate::allWriters() const
         }
     }
 
-    Q_FOREACH (const QString& externalPluginPath, externalPluginPaths) {
+    for (const QString& externalPluginPath : qAsConst(externalPluginPaths)) {
         ExternalWriter *plugin = new ExternalWriter(externalPluginPath);
-        Writer* writer = new Writer;
-        writer->d_ptr->m_plugin = plugin;
+        Writer writer;
+        writer.d->m_plugin = plugin;
+        writer.setAutoDeletePlugin(Writer::AutoDeletePlugin);
 
-        writers << writer;
+        m_allWriters.push_back(std::move(writer));
     }
 
-    return writers;
+    for (Writer& writer : m_allWriters) {
+        const QStringList lst = writer.mimetypes();
+        for (const QString& type : lst) {
+            m_writers.insert(type, &writer);
+        }
+    }
 }
 
 QList<Writer*> WriterCollection::fetchWriters(const QString& mimetype) const
 {
-    Q_D(const WriterCollection);
     QList<Writer*> plugins = d->m_writers.values(mimetype);
-    if (plugins.isEmpty()) {
-        auto it = d->m_writers.constBegin();
-        for (; it != d->m_writers.constEnd(); ++it) {
-            if (mimetype.startsWith(it.key()))
-                plugins << it.value();
+    if (!plugins.isEmpty()) {
+        return plugins;
+    }
+
+    // try to find the best matching more generic writer by mimetype inheritance
+    QMimeDatabase db;
+    auto type = db.mimeTypeForName(mimetype);
+    const QStringList ancestors = type.allAncestors();
+
+    for (const auto &ancestor : ancestors) {
+        if (ancestor == QLatin1String("application/octet-stream")) {
+            continue;
+        }
+        QList<Writer*> plugins = d->m_writers.values(ancestor);
+        if (!plugins.isEmpty()) {
+            qCDebug(KFILEMETADATA_LOG) << "Using inherited mimetype" << ancestor <<  "for" << mimetype;
+            return plugins;
         }
     }
 
